#include <boost/asio/steady_timer.hpp>
#include <boost/chrono.hpp>
#include "gtest_asio.h"

typedef boost::system::error_code error_code;

namespace ds {
namespace gtest {
namespace asio  {

::testing::AssertionResult runExpireFromNow(
    boost::asio::io_service& ioService,
    boost::chrono::milliseconds duration
)
{
    boost::asio::basic_waitable_timer<boost::chrono::steady_clock> timer(ioService);
    bool expired = false;

    timer.expires_from_now(duration);
    timer.async_wait([&](const error_code& e)
    {
        if (e) {
            return; //timer was aborted
        }
        ioService.stop();
        expired = true;
    });

    ioService.reset();
    ioService.run();
    if (expired) {
        return ::testing::AssertionFailure()
            << "io_service was not stopped within " << duration.count() << " ms"
        ;
    }
    return ::testing::AssertionSuccess()
        << "io_service was stopped before " << duration.count() << " ms"
    ;
}

::testing::AssertionResult runStop(boost::asio::io_service& ioService, boost::chrono::milliseconds duration)
{
    return runExpireFromNow(ioService, Timeout / 2);
}

::testing::AssertionResult runNoStop(boost::asio::io_service& ioService, boost::chrono::milliseconds duration)
{
    return !runExpireFromNow(ioService, Latency / 2);
}

::testing::AssertionResult runLatencyStop(boost::asio::io_service& ioService)
{
    auto x = runExpireFromNow(ioService, Latency / 2);
    if (x) {
        return x;
    }
    return runExpireFromNow(ioService, Latency);
}


}}} //namespace ds::gtest::asio
